#include "flatfile_pool.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include "wutils.h"
#include <string.h>
#include <strings.h>
#include <stdio.h>
#include <algorithm>
#include "stopos.h"

//
// this->status.extra is always pointing to
// the first free slot
//
// records are stored thusly:
// character to denote if this record exists or is deleted
// dbsep
// 
//   for a not deleted record, next fields:
//   key for this record
//   dbsep
//   1st parameter of put_record
//
//   for a deleted record:
//   pointer to next deleted record
//

// constructor
flatfile_pool::flatfile_pool()
{
}

std::string flatfile_pool::get_slot(const std::string &k)
{
  return k;
}

off_t flatfile_pool::strtoloc(const std::string &l)
{
  return this->reclen*stopos_key::KeyToNumber(l);
}

int flatfile_pool::get_record(std::string &r, const std::string &k, const std::string &l)
{
  int rc;

  //off_t p = this->reclen*stopos_key::KeyToNumber(k);
  off_t p = this->strtoloc(l);

  std::string rec;  // wwvv todo no need for rec

  rc = this->read_string(rec,p);
  if (rc !=0 )
    return rc;

  if (rec[0] == this->deleted)
    return this->NOTFOUND;

  // this record was not deleted, but it could be that
  // the key does not match
  //
  // find the key, it is between the first and second separator

  size_t q1 = 1;
  // find second separator. The first one is on postion 1
  size_t q2 = rec.find(dbsep,2);

  std::string key = rec.substr(q1+1,q2-q1-1);
  if (key != k)
    return this->NOTFOUND;

  rec.erase(0,q2+1);

  r = rec;

  return 0;
}

int flatfile_pool::put_record(const std::string &r, const std::string &k,const std::string &l)
{
  std::string buf;

  buf = buf + this->exists + dbsep + k + dbsep + r;

  //off_t p = this->reclen*stopos_key::KeyToNumber(slot);
  off_t p = this->strtoloc(l);

  return this->write_string(buf,p);
}

int flatfile_pool::remove_record(const stopos_key &k)
{
  int rc;

  off_t p = this->reclen*k.get<off_t>();

  std::string r;
  rc = this->read_string(r,p);
  if (rc != 0)
    return rc;

  r[0] = this->deleted;

  return this->write_string(r,p);
}

int flatfile_pool::read_string(std::string &r, off_t p)
{

  char b[this->reclen];
  off_t rc;
  rc = lseek(this->ff, p, SEEK_SET);
  if (rc == -1)
    return FETCHERROR;

  ssize_t rc1;
  rc1 = read(this->ff, b, this->reclen);
  if (rc1 != this->reclen)
    return FETCHERROR;

  r = b;

  return 0;
}


int flatfile_pool::write_string(const std::string &r, off_t p)
{
  int rc;
  char b[this->reclen];

  rc = lseek(this->ff,p,SEEK_SET);
  if (rc == -1)
    return STOREERROR;

  bzero(b, this->reclen);
  strncpy(b,r.c_str(),r.size());
  ssize_t first = 0;
  ssize_t togo  = this->reclen;
  ssize_t written;

  do
  {
    written = write(ff,&b[first],togo);
    if (written == -1)
    {
      perror("write:");
      return 2;
    }
    togo   -= written;
    first  += written;
  } while(togo !=0);

  return 0;
}


int flatfile_pool::create_db(const std::string &dbname)
{
  int rc; 

  rc = this->purge_db();
  if (rc != 0)
    return rc;

  // The db_name can be composed of parts separated with '/'
  // we try to create this path until the last '/'
  
  size_t p = dbname.find_last_of('/');
  rc = mkdir_p(dbname.substr(0,p+1),0755);
  if (rc != 0)
    return CREATEERROR;

  rc = this->open_db();
  if (rc != 0)
    return rc;
  this->close_db();

  this->status.extra = NumberToString(0);

  return 0;
}

int flatfile_pool::purge_db(void)
{
  if (this->db_name == 0)
    return NOFILENAME;

  this->close_db();

  unlink(this->db_name);
  return 0;
}

int flatfile_pool::open_db(void)
{
  int rc;

  if (this->db_name == 0)
    return NOFILENAME;

  if (this->db_open)
    return 0;

  this->ff = open(this->db_name, O_CREAT|O_RDWR,0644);
  if (this->ff < 0)
    return OPENERROR;

  struct flock lock;
  lock.l_type   = F_RDLCK|F_WRLCK;
  lock.l_whence = SEEK_SET;
  lock.l_start  = 0;
  lock.l_len    = 0;

  rc = fcntl(this->ff,F_SETLKW,&lock);
  if (rc < 0)
    return LOCKERROR;

  this->db_open = 1;

  return 0;
}

int flatfile_pool::close_db(void)
{
  if (!this->db_open)
     return 0;

  int rc;

  struct flock lock;
  lock.l_type   = F_UNLCK;
  lock.l_whence = SEEK_SET;
  lock.l_start  = 0;
  lock.l_len    = 0;

  rc = fcntl(this->ff,F_SETLKW,&lock);
  if (rc < 0)
    return UNLOCKERROR;

  rc = close(this->ff);
  if (rc == -1)
    return CLOSEERROR;

  this->db_open = 0;

  return 0;
}

void flatfile_pool::dump_db(void)
{
  std::string r;
  off_t n = 1;

  while (read_string(r,n*this->reclen)==0)
  {
    std::cerr << r << std::endl;
    n++;
  }
}
